
package main

import (
	"cycoreutils"
	"fmt"

	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/core/chaincode/lib/cid"
	"github.com/pkg/errors"
)

var ApplicationLogger = shim.NewLogger("application")

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Record Application to the ledger
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func recordApplication(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var err error
	var collectionName string
	var Avalbytes []byte
	var Application = &application{}

	// Convert the arg to a recordApplication object
	ApplicationLogger.Info("recordApplication() : Arguments for recordApplication : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Application)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Application object")).Error(), nil)
	}

	Application.ObjectType = "Application"

	callerId, err := cid.GetID(stub)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to get transaction caller ID")).Error(), nil)
	}

	Application.OwnerID = callerId

	// Query and Retrieve the Full recordApplication
	keys := []string{  Application.ApplicationID  }
	ApplicationLogger.Debug("Keys for Application %s: ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Application.ObjectType, keys, collectionName)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Application object")).Error(), nil)
	}

	if Avalbytes != nil {
		return cycoreutils.ConstructResponse("SASTQRY008E", fmt.Sprintf("Application already exists"), nil)
	}

	ApplicationBytes, _ := cycoreutils.ObjecttoJSON(Application)

	err = cycoreutils.UpdateObject(stub, Application.ObjectType, keys, ApplicationBytes,collectionName)

	if err != nil {
		ApplicationLogger.Errorf("recordApplication() : Error inserting Application object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Application object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTREC011S", fmt.Sprintf("Successfully Recorded Application object"), ApplicationBytes)
}

//////////////////////////////////////////////////////////////
/// Query Application Info from the ledger
//////////////////////////////////////////////////////////////
func queryApplication(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var err error
	var Avalbytes []byte
	var collectionName string
	var Application = &application{}

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Application ID}. Received %d arguments", len(args)), nil)
	}

	Application.ObjectType = "Application"

	err = cycoreutils.JSONtoObject([]byte(args[0]), Application)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrap(err, "Failed to convert arg[0] to Application object")).Error(), nil)
	}

	// Query and Retrieve the Full Application
	keys := []string{  Application.ApplicationID  }
	ApplicationLogger.Info("Keys for Application : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Application.ObjectType, keys, collectionName)

	if err != nil {
		cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Application object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Application not found"), nil)
	}

	err = cycoreutils.JSONtoObject(Avalbytes, Application)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert query result to Application object")).Error(), nil)
	}

	ApplicationLogger.Info("queryApplication() : Returning Application results")

	return cycoreutils.ConstructResponse("SASTQRY005S", fmt.Sprintf("Successfully Queried Application object"), Avalbytes)
}

//////////////////////////////////////////////////////////////
/// Query Application List from the ledger
//////////////////////////////////////////////////////////////
func queryApplicationList(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var collectionName string
	collectionName = ""

	queryString := fmt.Sprintf("{\"selector\":{\"objectType\":\"Application\"}}")
	ApplicationLogger.Info("Query List: queryString: ", queryString)

	queryResults, err := cycoreutils.GetQueryResultForQueryString(stub, queryString,collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY006E", (errors.Wrapf(err, "Failed to query Application object list")).Error(), nil)
	}

	if queryResults == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("No Application record found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY007S", fmt.Sprintf("Successfully Retrieved the list of Application objects "), queryResults)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Update Application to the ledger by getting a copy of it, updating that copy, and overwriting the original to a newer version
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func updateApplication(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
    var err error
	var Avalbytes []byte
	var collectionName string

	var Application = &application{}
	// Convert the arg to a updateApplication object
	ApplicationLogger.Info("updateApplication() : Arguments for Query: Application : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Application)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Application object")).Error(), nil)
	}
	Application.ObjectType = "Application"
	// Query and Retrieve the Full updateApplication
	keys := []string{  Application.ApplicationID  }
	ApplicationLogger.Info("Keys for Application : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Application.ObjectType, keys, collectionName)


	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Application object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Application does not exist to update"), nil)
	}

	var currentProps = &application{}
	err = cycoreutils.JSONtoObject(Avalbytes, currentProps)

	callerId, err := cid.GetID(stub)

	if callerId != currentProps.OwnerID {
		return cycoreutils.ConstructResponse("SASTUPD009E", fmt.Sprintf("caller is not asset owner"), nil)
	}

	Application.OwnerID = currentProps.OwnerID
	ApplicationBytes, _ := cycoreutils.ObjecttoJSON(Application)
	err = cycoreutils.UpdateObject(stub, Application.ObjectType, keys, ApplicationBytes, collectionName)

	if err != nil {
		ApplicationLogger.Errorf("updateApplication() : Error updating Application object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Application object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTUPD010S", fmt.Sprintf("Successfully Updated Application object"), ApplicationBytes)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Delete Application from the ledger.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func deleteApplication(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var err error
	ObjectType := "Application"
	var collectionName string
	var Application = &application{}

	collectionName = ""

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Application ID}. Received %d arguments", len(args)), nil)
	}
	// Convert the arg to a deleteApplication object
	ApplicationLogger.Info("deleteApplication() : Arguments for Query: Application : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Application)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Application object")).Error(), nil)
	}

	// Query and Retrieve the Full deleteApplication
	keys := []string{  Application.ApplicationID  }
	ApplicationLogger.Info("Keys for Application : ", keys)

	err = cycoreutils.DeleteObject(stub, ObjectType, keys,collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTDEL012E", (errors.Wrapf(err, "Failed to delete Application object")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTDEL013I", fmt.Sprintf("Successfully Deleted Application object"), nil)
}

//////////////////////////////////////////////////////////////
/// Query Application History from the ledger
//////////////////////////////////////////////////////////////
func queryApplicationHistory(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
    	var err error
	var Avalbytes []byte
	var Application = &application{}

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Application ID}. Received %d arguments", len(args)), nil)
	}

	err = cycoreutils.JSONtoObject([]byte(args[0]), Application)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Application object")).Error(), nil)
	}

	// Query and Retrieve the Full Application
	keys := []string{  Application.ApplicationID  }
	ApplicationLogger.Info("Keys for Application : ", keys)

    Avalbytes, err = cycoreutils.GetObjectHistory(stub, "Application", keys)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Application object history")).Error(), nil)
	}
	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Application object not found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY014S", fmt.Sprintf("Successfully Queried Application object History"), Avalbytes)
}
