
package main

import (
	"cycoreutils"
	"fmt"

	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/core/chaincode/lib/cid"
	"github.com/pkg/errors"
)

var FunctionLogger = shim.NewLogger("function")

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Record Function to the ledger
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func recordFunction(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var err error
	var collectionName string
	var Avalbytes []byte
	var Function = &function{}

	// Convert the arg to a recordFunction object
	FunctionLogger.Info("recordFunction() : Arguments for recordFunction : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Function)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Function object")).Error(), nil)
	}

	Function.ObjectType = "Function"

	callerId, err := cid.GetID(stub)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to get transaction caller ID")).Error(), nil)
	}

	Function.OwnerID = callerId

	// Query and Retrieve the Full recordFunction
	keys := []string{  Function.FunctionID  }
	FunctionLogger.Debug("Keys for Function %s: ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Function.ObjectType, keys, collectionName)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Function object")).Error(), nil)
	}

	if Avalbytes != nil {
		return cycoreutils.ConstructResponse("SASTQRY008E", fmt.Sprintf("Function already exists"), nil)
	}

	FunctionBytes, _ := cycoreutils.ObjecttoJSON(Function)

	err = cycoreutils.UpdateObject(stub, Function.ObjectType, keys, FunctionBytes,collectionName)

	if err != nil {
		FunctionLogger.Errorf("recordFunction() : Error inserting Function object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Function object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTREC011S", fmt.Sprintf("Successfully Recorded Function object"), FunctionBytes)
}

//////////////////////////////////////////////////////////////
/// Query Function Info from the ledger
//////////////////////////////////////////////////////////////
func queryFunction(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var err error
	var Avalbytes []byte
	var collectionName string
	var Function = &function{}

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Function ID}. Received %d arguments", len(args)), nil)
	}

	Function.ObjectType = "Function"

	err = cycoreutils.JSONtoObject([]byte(args[0]), Function)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrap(err, "Failed to convert arg[0] to Function object")).Error(), nil)
	}

	// Query and Retrieve the Full Function
	keys := []string{  Function.FunctionID  }
	FunctionLogger.Info("Keys for Function : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Function.ObjectType, keys, collectionName)

	if err != nil {
		cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Function object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Function not found"), nil)
	}

	err = cycoreutils.JSONtoObject(Avalbytes, Function)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert query result to Function object")).Error(), nil)
	}

	FunctionLogger.Info("queryFunction() : Returning Function results")

	return cycoreutils.ConstructResponse("SASTQRY005S", fmt.Sprintf("Successfully Queried Function object"), Avalbytes)
}

//////////////////////////////////////////////////////////////
/// Query Function List from the ledger
//////////////////////////////////////////////////////////////
func queryFunctionList(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var collectionName string
	collectionName = ""

	queryString := fmt.Sprintf("{\"selector\":{\"objectType\":\"Function\"}}")
	FunctionLogger.Info("Query List: queryString: ", queryString)

	queryResults, err := cycoreutils.GetQueryResultForQueryString(stub, queryString,collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY006E", (errors.Wrapf(err, "Failed to query Function object list")).Error(), nil)
	}

	if queryResults == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("No Function record found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY007S", fmt.Sprintf("Successfully Retrieved the list of Function objects "), queryResults)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Update Function to the ledger by getting a copy of it, updating that copy, and overwriting the original to a newer version
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func updateFunction(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
  var err error
	var Avalbytes []byte
	var collectionName string

	var Function = &function{}
	// Convert the arg to a updateFunction object
	FunctionLogger.Info("updateFunction() : Arguments for Query: Function : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Function)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Function object")).Error(), nil)
	}
	Function.ObjectType = "Function"
	// Query and Retrieve the Full updateFunction
	keys := []string{  Function.FunctionID  }
	FunctionLogger.Info("Keys for Function : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Function.ObjectType, keys, collectionName)


	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Function object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Function does not exist to update"), nil)
	}

	var currentProps = &function{}
	err = cycoreutils.JSONtoObject(Avalbytes, currentProps)

	callerId, err := cid.GetID(stub)

	if callerId != currentProps.OwnerID {
		return cycoreutils.ConstructResponse("SASTUPD009E", fmt.Sprintf("caller is not asset owner"), nil)
	}

	Function.OwnerID = currentProps.OwnerID
	FunctionBytes, _ := cycoreutils.ObjecttoJSON(Function)

	err = cycoreutils.UpdateObject(stub, Function.ObjectType, keys, FunctionBytes, collectionName)

	if err != nil {
		FunctionLogger.Errorf("updateFunction() : Error updating Function object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Function object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTUPD010S", fmt.Sprintf("Successfully Updated Function object"), FunctionBytes)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Delete Function from the ledger.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func deleteFunction(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var err error
	ObjectType := "Function"
	var collectionName string
	var Function = &function{}

	collectionName = ""

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Function ID}. Received %d arguments", len(args)), nil)
	}
	// Convert the arg to a deleteFunction object
	FunctionLogger.Info("deleteFunction() : Arguments for Query: Function : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Function)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Function object")).Error(), nil)
	}

	// Query and Retrieve the Full deleteFunction
	keys := []string{  Function.FunctionID  }
	FunctionLogger.Info("Keys for Function : ", keys)

	err = cycoreutils.DeleteObject(stub, ObjectType, keys,collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTDEL012E", (errors.Wrapf(err, "Failed to delete Function object")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTDEL013I", fmt.Sprintf("Successfully Deleted Function object"), nil)
}

//////////////////////////////////////////////////////////////
/// Query Function History from the ledger
//////////////////////////////////////////////////////////////
func queryFunctionHistory(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
    	var err error
	var Avalbytes []byte
	var Function = &function{}

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Function ID}. Received %d arguments", len(args)), nil)
	}

	err = cycoreutils.JSONtoObject([]byte(args[0]), Function)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Function object")).Error(), nil)
	}

	// Query and Retrieve the Full Function
	keys := []string{  Function.FunctionID  }
	FunctionLogger.Info("Keys for Function : ", keys)

    Avalbytes, err = cycoreutils.GetObjectHistory(stub, "Function", keys)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Function object history")).Error(), nil)
	}
	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Function object not found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY014S", fmt.Sprintf("Successfully Queried Function object History"), Avalbytes)
}
