
package main

import (
	"cycoreutils"
	"fmt"

	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/core/chaincode/lib/cid"
	"github.com/pkg/errors"
)

var NodeLogger = shim.NewLogger("node")

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Record Node to the ledger
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func recordNode(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var err error
	var collectionName string
	var Avalbytes []byte
	var Node = &node{}

	// Convert the arg to a recordNode object
	NodeLogger.Info("recordNode() : Arguments for recordNode : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Node)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Node object")).Error(), nil)
	}

	Node.ObjectType = "Node"

	callerId, err := cid.GetID(stub)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to get transaction caller ID")).Error(), nil)
	}

	Node.OwnerID = callerId

	// Query and Retrieve the Full recordNode
	keys := []string{  Node.NodeID  }
	NodeLogger.Debug("Keys for Node %s: ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Node.ObjectType, keys, collectionName)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Node object")).Error(), nil)
	}

	if Avalbytes != nil {
		return cycoreutils.ConstructResponse("SASTQRY008E", fmt.Sprintf("Node already exists"), nil)
	}

	NodeBytes, _ := cycoreutils.ObjecttoJSON(Node)

	err = cycoreutils.UpdateObject(stub, Node.ObjectType, keys, NodeBytes,collectionName)

	if err != nil {
		NodeLogger.Errorf("recordNode() : Error inserting Node object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Node object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTREC011S", fmt.Sprintf("Successfully Recorded Node object"), NodeBytes)
}

//////////////////////////////////////////////////////////////
/// Query Node Info from the ledger
//////////////////////////////////////////////////////////////
func queryNode(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var err error
	var Avalbytes []byte
	var collectionName string
	var Node = &node{}

	Node.ObjectType = "Node"

	err = cycoreutils.JSONtoObject([]byte(args[0]), Node)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrap(err, "Failed to convert arg[0] to Node object")).Error(), nil)
	}

	// Query and Retrieve the Full Node
	keys := []string{  Node.NodeID  }
	NodeLogger.Info("Keys for Node : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Node.ObjectType, keys, collectionName)

	if err != nil {
		cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Node object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Node not found"), nil)
	}

	err = cycoreutils.JSONtoObject(Avalbytes, Node)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert query result to Node object")).Error(), nil)
	}

	NodeLogger.Info("queryNode() : Returning Node results")

	return cycoreutils.ConstructResponse("SASTQRY005S", fmt.Sprintf("Successfully Queried Node object"), Avalbytes)
}

//////////////////////////////////////////////////////////////
/// Query Node List from the ledger
//////////////////////////////////////////////////////////////
func queryNodeList(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var collectionName string
	collectionName = ""

	queryString := fmt.Sprintf("{\"selector\":{\"objectType\":\"Node\"}}")
	NodeLogger.Info("Query List: queryString: ", queryString)

	queryResults, err := cycoreutils.GetQueryResultForQueryString(stub, queryString,collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY006E", (errors.Wrapf(err, "Failed to query Node object list")).Error(), nil)
	}

	if queryResults == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("No Node record found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY007S", fmt.Sprintf("Successfully Retrieved the list of Node objects "), queryResults)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Update Node to the ledger by getting a copy of it, updating that copy, and overwriting the original to a newer version
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func updateNode(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
  var err error
	var Avalbytes []byte
	var collectionName string
	var Node = &node{}
	// Convert the arg to a updateNode object
	NodeLogger.Info("updateNode() : Arguments for Query: Node : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Node)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Node object")).Error(), nil)
	}
	Node.ObjectType = "Node"
	// Query and Retrieve the Full updateNode
	keys := []string{  Node.NodeID  }
	NodeLogger.Info("Keys for Node : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Node.ObjectType, keys, collectionName)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Node object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Node does not exist to update"), nil)
	}

	var currentProps = &node{}
	err = cycoreutils.JSONtoObject(Avalbytes, currentProps)

	callerId, err := cid.GetID(stub)

	if callerId != currentProps.OwnerID {
		return cycoreutils.ConstructResponse("SASTUPD009E", fmt.Sprintf("caller is not asset owner"), nil)
	}

	Node.OwnerID = currentProps.OwnerID
	NodeBytes, _ := cycoreutils.ObjecttoJSON(Node)

	err = cycoreutils.UpdateObject(stub, Node.ObjectType, keys, NodeBytes, collectionName)

	if err != nil {
		NodeLogger.Errorf("updateNode() : Error updating Node object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Node object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTUPD010S", fmt.Sprintf("Successfully Updated Node object"), NodeBytes)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Delete Node from the ledger.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func deleteNode(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var err error
	ObjectType := "Node"
	var collectionName string
	var Node = &node{}

	collectionName = ""

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Node ID}. Received %d arguments", len(args)), nil)
	}
	// Convert the arg to a deleteNode object
	NodeLogger.Info("deleteNode() : Arguments for Query: Node : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Node)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Node object")).Error(), nil)
	}

	// Query and Retrieve the Full deleteNode
	keys := []string{  Node.NodeID  }
	NodeLogger.Info("Keys for Node : ", keys)

	err = cycoreutils.DeleteObject(stub, ObjectType, keys,collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTDEL012E", (errors.Wrapf(err, "Failed to delete Node object")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTDEL013I", fmt.Sprintf("Successfully Deleted Node object"), nil)
}

//////////////////////////////////////////////////////////////
/// Query Node History from the ledger
//////////////////////////////////////////////////////////////
func queryNodeHistory(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
    	var err error
	var Avalbytes []byte
	var Node = &node{}

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Node ID}. Received %d arguments", len(args)), nil)
	}

	err = cycoreutils.JSONtoObject([]byte(args[0]), Node)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Node object")).Error(), nil)
	}

	// Query and Retrieve the Full Node
	keys := []string{  Node.NodeID  }
	NodeLogger.Info("Keys for Node : ", keys)

    Avalbytes, err = cycoreutils.GetObjectHistory(stub, "Node", keys)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Node object history")).Error(), nil)
	}
	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Node object not found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY014S", fmt.Sprintf("Successfully Queried Node object History"), Avalbytes)
}
