
package main

import (
	"cycoreutils"
	"fmt"
	"strconv"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/pkg/errors"
)

var WalletLogger = shim.NewLogger("wallet")

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Record Wallet to the ledger
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func recordWallet(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var err error
	var collectionName string
	var Avalbytes []byte
	var Wallet = &wallet{}

	// Convert the arg to a recordWallet object
	WalletLogger.Info("recordWallet() : Arguments for recordWallet : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Wallet)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Wallet object")).Error(), nil)
	}

	Wallet.ObjectType = "Wallet"

	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to get transaction caller ID")).Error(), nil)
	}

	// Query and Retrieve the Full recordWallet
	keys := []string{  Wallet.WalletID  }
	WalletLogger.Debug("Keys for Wallet %s: ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Wallet.ObjectType, keys, collectionName)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Wallet object")).Error(), nil)
	}

	if Avalbytes != nil {
		return cycoreutils.ConstructResponse("SASTQRY008E", fmt.Sprintf("Wallet already exists"), nil)
	}

	WalletBytes, _ := cycoreutils.ObjecttoJSON(Wallet)

	err = cycoreutils.UpdateObject(stub, Wallet.ObjectType, keys, WalletBytes, collectionName)


	if err != nil {
		WalletLogger.Errorf("recordWallet() : Error inserting Wallet object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Wallet object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTREC011S", fmt.Sprintf("Successfully Recorded Wallet object"), WalletBytes)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Record Transaction to the ledger
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func _recordTransaction(stub shim.ChaincodeStubInterface, args []string) bool {
	var err error
	var collectionName string
	var Avalbytes []byte
	var Transaction = &transaction{}

	// Convert the arg to a recordTransaction object
	WalletLogger.Info("recordTransaction() : Arguments for recordTransaction : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), Transaction)
	if err != nil {
		return false
	}

	Transaction.ObjectType = "Transaction"

	// Query and Retrieve the Full recordTransaction
	keys := []string{  Transaction.TransactionID  }
	WalletLogger.Debug("Keys for Transaction %s: ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Transaction.ObjectType, keys, collectionName)

	if err != nil {
		return false
	}

	if Avalbytes != nil {
		return false
	}

	TransactionBytes, _ := cycoreutils.ObjecttoJSON(Transaction)

	err = cycoreutils.UpdateObject(stub, Transaction.ObjectType, keys, TransactionBytes, collectionName)


	if err != nil {
		WalletLogger.Errorf("recordTransaction() : Error inserting Transaction object into LedgerState %s", err)
		return false
	}

	return true
}

//////////////////////////////////////////////////////////////
/// Query Wallet Info from the ledger
//////////////////////////////////////////////////////////////
func queryWallet(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var err error
	var Avalbytes []byte
	var collectionName string
	var Wallet = &wallet{}

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Wallet ID}. Received %d arguments", len(args)), nil)
	}

	Wallet.ObjectType = "Wallet"

	err = cycoreutils.JSONtoObject([]byte(args[0]), Wallet)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrap(err, "Failed to convert arg[0] to Wallet object")).Error(), nil)
	}

	// Query and Retrieve the Full Wallet
	keys := []string{  Wallet.WalletID  }
	WalletLogger.Info("Keys for Wallet : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, Wallet.ObjectType, keys, collectionName)

	if err != nil {
		cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Wallet object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Wallet not found"), nil)
	}

	err = cycoreutils.JSONtoObject(Avalbytes, Wallet)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert query result to Wallet object")).Error(), nil)
	}

	WalletLogger.Info("queryWallet() : Returning Wallet results")

	return cycoreutils.ConstructResponse("SASTQRY005S", fmt.Sprintf("Successfully Queried Wallet object"), Avalbytes)
}

//////////////////////////////////////////////////////////////
/// Query Wallet List from the ledger
//////////////////////////////////////////////////////////////
func queryWalletList(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {

	var collectionName string
	collectionName = ""

	queryString := fmt.Sprintf("{\"selector\":{\"objectType\":\"Wallet\"}}")
	WalletLogger.Info("Query List: queryString: ", queryString)

	queryResults, err := cycoreutils.GetQueryResultForQueryString(stub, queryString,collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY006E", (errors.Wrapf(err, "Failed to query Wallet object list")).Error(), nil)
	}

	if queryResults == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("No Wallet record found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY007S", fmt.Sprintf("Successfully Retrieved the list of Wallet objects "), queryResults)
}

//////////////////////////////////////////////////////////////
/// Query Transaction List from the ledger
//////////////////////////////////////////////////////////////
func queryTransactionList(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var collectionName string
	collectionName = ""

	queryString := fmt.Sprintf("{\"selector\":{\"objectType\":\"Transaction\"}}")

	queryResults, err := cycoreutils.GetQueryResultForQueryString(stub, queryString, collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY006E", (errors.Wrapf(err, "Failed to query Transaction object list")).Error(), nil)
	}

	if queryResults == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("No Transaction record found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY007S", fmt.Sprintf("Successfully Retrieved the list of Transaction objects "), queryResults)
}

//////////////////////////////////////////////////////////////
/// Query Transaction by Type
//////////////////////////////////////////////////////////////
func queryTransactionByType(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var err error
	var collectionName string
	collectionName = ""

	var Transaction = &transaction{}
	err = cycoreutils.JSONtoObject([]byte(args[0]), Transaction)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrap(err, "Failed to convert arg[0] to Wallet object")).Error(), nil)
	}

	queryString := fmt.Sprintf("{\"selector\":{\"objectType\":\"Transaction\",\"transactionType\":\"" + Transaction.TransactionType + "\"}}")

	queryResults, err := cycoreutils.GetQueryResultForQueryString(stub, queryString, collectionName)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY006E", (errors.Wrapf(err, "Failed to query Transaction object list")).Error(), nil)
	}

	if queryResults == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("No Transaction record found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY007S", fmt.Sprintf("Successfully Retrieved the list of Transaction objects "), queryResults)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// tokenTransfer external function
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func tokenTransfer(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var err error
	var fromWalletBytes []byte
	var toWalletBytes []byte
	var collectionName string
	var transaction = &transaction{}


	transactionRecorded := _recordTransaction(stub, args)

	if !transactionRecorded {
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Failed to record transaction")).Error(), nil)
	}

	// Convert the arg to a tokenTransfer object
	WalletLogger.Info("tokenTransfer() : Arguments for Query: Wallet: ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), transaction)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Wallet object")).Error(), nil)
	}

	fromKeys := []string{  transaction.FromWalletID  }
	WalletLogger.Info("Keys for from Wallet : ", fromKeys)

	toKeys := []string{  transaction.ToWalletID  }
	WalletLogger.Info("Keys for to Wallet : ", toKeys)

	collectionName = ""

	fmt.Sprintf("transaction.FromWalletID = %s,  transaction.ToWalletID = %s", transaction.FromWalletID, transaction.ToWalletID)

	fromWalletBytes, err = cycoreutils.QueryObject(stub, "Wallet", fromKeys, collectionName)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query fromWallet object")).Error(), nil)
	}

	if fromWalletBytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("fromWallet does not exist"), nil)
	}

	var fromWallet = &wallet{}
	err = cycoreutils.JSONtoObject(fromWalletBytes, fromWallet)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to get from wallet current props")).Error(), nil)
	}

	toWalletBytes, err = cycoreutils.QueryObject(stub, "Wallet", toKeys, collectionName)


	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query toWallet object")).Error(), nil)
	}

	if toWalletBytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("to wallet does not exist"), nil)
	}

	var toWallet = &wallet{}
	err = cycoreutils.JSONtoObject(toWalletBytes, toWallet)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to get to wallet current props")).Error(), nil)
	}

	if fromWallet.WalletID == toWallet.WalletID {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("cannot transfer token to self"), nil)
	}

	fromWalletBalance, _ := strconv.Atoi(fromWallet.Balance)
	transactionAmount, _ := strconv.Atoi(transaction.Amount)

	if fromWalletBalance < transactionAmount {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("from wallet does not have enough balance"), nil)
	}

	fromWallet.Balance = strconv.Itoa(fromWalletBalance - transactionAmount)

	toWalletBalance, _ := strconv.Atoi(toWallet.Balance)
	toWallet.Balance = strconv.Itoa(toWalletBalance + transactionAmount)

	fromWalletBytes, _ = cycoreutils.ObjecttoJSON(fromWallet)

	toWalletBytes, _ = cycoreutils.ObjecttoJSON(toWallet)


	err = cycoreutils.UpdateObject(stub, "Wallet", fromKeys, fromWalletBytes, collectionName)

	err = cycoreutils.UpdateObject(stub, "Wallet", toKeys, toWalletBytes, collectionName)


	if err != nil {
		WalletLogger.Errorf("tokenTransfer() : Error updating Wallet object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Wallet object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTUPD010S", fmt.Sprintf("Successfully Updated Wallet object"), toWalletBytes)

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// _tokenTransfer internal function
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func _tokenTransfer(stub shim.ChaincodeStubInterface, fromWalletID string, toWalletID string, amount int) bool {
	var err error
	var fromWalletBytes []byte
	var toWalletBytes []byte

	fromKeys := []string{  fromWalletID  }

	toKeys := []string{  toWalletID  }

	// WalletLogger.Info("fromWalletID = %s, toWalletID = %s, amount = %d", fromWalletID, toWalletID, amount)

	fromWalletBytes, err = cycoreutils.QueryObject(stub, "Wallet", fromKeys, "")

	if err != nil {
		return false
	}

	if fromWalletBytes == nil {
		return false
	}

	var fromWallet = &wallet{}
	err = cycoreutils.JSONtoObject(fromWalletBytes, fromWallet)

	if err != nil {
		return false
	}

	toWalletBytes, err = cycoreutils.QueryObject(stub, "Wallet", toKeys, "")


	if err != nil {
		return false
	}

	if toWalletBytes == nil {
		return false
	}

	var toWallet = &wallet{}
	err = cycoreutils.JSONtoObject(toWalletBytes, toWallet)

	if err != nil {
		return false
	}

	if fromWallet.WalletID == toWallet.WalletID {
		return false
	}

	fromWalletBalance, _ := strconv.Atoi(fromWallet.Balance)
	toWalletBalance, _ := strconv.Atoi(toWallet.Balance)

	if fromWalletBalance < amount {
		return false
	}

	WalletLogger.Info("BEFORE!!: fromWallet.Balance = %d, toWallet.Balance %d", fromWallet.Balance, toWallet.Balance)

	fromWallet.Balance = strconv.Itoa(fromWalletBalance - amount)
	toWallet.Balance = strconv.Itoa(toWalletBalance + amount)

	// WalletLogger.Info("AFTER!!: fromWallet.Balance = %d, toWallet.Balance %d", fromWallet.Balance, toWallet.Balance)

	fromWalletBytes, _ = cycoreutils.ObjecttoJSON(fromWallet)

	toWalletBytes, _ = cycoreutils.ObjecttoJSON(toWallet)

	cycoreutils.UpdateObject(stub, "Wallet", fromKeys, fromWalletBytes, "")

	cycoreutils.UpdateObject(stub, "Wallet", toKeys, toWalletBytes, "")

	return true
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// mint tokens to the wallet
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func tokenMint(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
  var err error
	var Avalbytes []byte
	var collectionName string

	var transaction = &transaction{}

	transactionRecorded := _recordTransaction(stub, args)

	if !transactionRecorded {
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Failed to record transaction")).Error(), nil)
	}

	transaction.FromWalletID = ""

	// Convert the arg to a tokenTransfer object
	WalletLogger.Info("tokenTransfer() : Arguments for Query: Wallet : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), transaction)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Wallet object")).Error(), nil)
	}

	// Query and Retrieve the Full tokenTransfer
	keys := []string{  transaction.ToWalletID  }
	WalletLogger.Info("Keys for Wallet : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, "Wallet", keys, collectionName)


	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Wallet object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Wallet does not exist to update, txn ToWalletID %s, transaction.Amount %d, args[0] %s", transaction.ToWalletID, transaction.Amount, args[0]), nil)
	}

	var currentProps = &wallet{}
	err = cycoreutils.JSONtoObject(Avalbytes, currentProps)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to get wallet current props")).Error(), nil)
	}

	currentProps.Balance = currentProps.Balance + transaction.Amount

	updatedBytes, _ := cycoreutils.ObjecttoJSON(currentProps)

	err = cycoreutils.UpdateObject(stub, "Wallet", keys, updatedBytes, collectionName)

	if err != nil {
		WalletLogger.Errorf("tokenTransfer() : Error updating Wallet object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Wallet object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTUPD010S", fmt.Sprintf("Successfully Updated Wallet object"), updatedBytes)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// burn tokens to the wallet
////////////////////////////////////////////////////////////////////////////////////////////////////////////
func tokenBurn(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
	var err error
	var Avalbytes []byte
	var collectionName string

	var transaction = &transaction{}

	transactionRecorded := _recordTransaction(stub, args)

	if !transactionRecorded {
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Failed to record transaction")).Error(), nil)
	}

	transaction.ToWalletID = ""

	// Convert the arg to a tokenTransfer object
	WalletLogger.Info("tokenTransfer() : Arguments for Query: Wallet : ", args[0])
	err = cycoreutils.JSONtoObject([]byte(args[0]), transaction)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Wallet object")).Error(), nil)
	}

	// Query and Retrieve the Full tokenTransfer
	keys := []string{  transaction.FromWalletID  }
	WalletLogger.Info("Keys for Wallet : ", keys)

	collectionName = ""

	Avalbytes, err = cycoreutils.QueryObject(stub, "Wallet", keys, collectionName)


	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Wallet object")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Wallet does not exist to update"), nil)
	}

	var currentProps = &wallet{}
	err = cycoreutils.JSONtoObject(Avalbytes, currentProps)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to get wallet current props")).Error(), nil)
	}

	currentBalance, _ := strconv.Atoi(currentProps.Balance)
	transactionAmount, _ := strconv.Atoi(transaction.Amount)

	if transactionAmount > currentBalance {
		return cycoreutils.ConstructResponse("SASTUPD009E", fmt.Sprintf("Token burn failed, wallet does not have enough balance"), nil)
	}

	currentProps.Balance = strconv.Itoa(currentBalance - transactionAmount)

	updatedBytes, _ := cycoreutils.ObjecttoJSON(currentProps)

	err = cycoreutils.UpdateObject(stub, "Wallet", keys, updatedBytes, collectionName)

	if err != nil {
		WalletLogger.Errorf("tokenTransfer() : Error updating Wallet object into LedgerState %s", err)
		return cycoreutils.ConstructResponse("SASTUPD009E", (errors.Wrapf(err, "Wallet object update failed")).Error(), nil)
	}

	return cycoreutils.ConstructResponse("SASTUPD010S", fmt.Sprintf("Successfully Updated Wallet object"), updatedBytes)
}

//////////////////////////////////////////////////////////////
/// Query Wallet History from the ledger
//////////////////////////////////////////////////////////////
func queryWalletHistory(stub shim.ChaincodeStubInterface, args []string) cycoreutils.Response {
    	var err error
	var Avalbytes []byte
	var Wallet = &wallet{}

	// In future , it should be > 1 and ,= mo_of_keys for object
	if len(args) != 1 {
		return cycoreutils.ConstructResponse("SUSRPARM001E", fmt.Sprintf("Expecting Wallet ID}. Received %d arguments", len(args)), nil)
	}

	err = cycoreutils.JSONtoObject([]byte(args[0]), Wallet)
	if err != nil {
		return cycoreutils.ConstructResponse("SASTCONV002E", (errors.Wrapf(err, "Failed to convert arg[0] to Wallet object")).Error(), nil)
	}

	// Query and Retrieve the Full Wallet
	keys := []string{  Wallet.WalletID  }
	WalletLogger.Info("Keys for Wallet : ", keys)

  Avalbytes, err = cycoreutils.GetObjectHistory(stub, "Wallet", keys)

	if err != nil {
		return cycoreutils.ConstructResponse("SASTQRY003E", (errors.Wrapf(err, "Failed to query Wallet object history")).Error(), nil)
	}

	if Avalbytes == nil {
		return cycoreutils.ConstructResponse("SASTDNE004E", fmt.Sprintf("Wallet object not found"), nil)
	}

	return cycoreutils.ConstructResponse("SASTQRY014S", fmt.Sprintf("Successfully Queried Wallet object History"), Avalbytes)
}

//////////////////////////////////////////////////////////////
/// make wallet id from owner ID
//////////////////////////////////////////////////////////////
func makeWalletID(OwnerID string) string {
	// TODO: for now just use OwnerID, should use some sort of convertion function
	return OwnerID
}
