const axios = require("axios");
const chai = require("chai");
const chaiAsPromised = require("chai-as-promised");
chai.use(chaiAsPromised);
const expect = chai.expect;
const _ = require("lodash");
const nanoid = require("nanoid");
const log4js = require("log4js");
const logger = log4js.getLogger("Function-test");

const expectThrowsAsync = async (method, errorMessage) => {
  let error = null;
  try {
    await method();
  } catch (err) {
    error = err;
  }
  expect(error).to.be.an("Error");
  if (errorMessage) {
    expect(error.message).to.equal(errorMessage);
  }
};

describe("Wallet api ", async function() {
  this.timeout(45000);
  let xAccessToken;
  let allWallets;
  let firstWalletId;

  afterEach(function(done) {
    setTimeout(function() {
      done();
    }, 30000);
  });

  describe("Admin actions", async function() {
    before(async () => {
      try {
        const email = `function${nanoid()}@fakeemail.com;`;
        await axios({
          method: "post",
          url: "http://localhost:3000/users/register",
          data: {
            email,
            password: "p@ssw0rd",
            org: "org1",
            role: "admin"
          },
          headers: {
            "content-type": "application/json"
          }
        });
        let response = await axios({
          method: "post",
          url: "http://localhost:3000/users/login",
          data: {
            email,
            password: "p@ssw0rd",
            org: "org1"
          },
          headers: {
            "content-type": "application/json"
          }
        });
        xAccessToken = response.data.accessToken;
        allWallets = await axios({
          method: "get",
          url: "http://localhost:3000/wallets",
          headers: {
            "content-type": "application/json",
            "x-access-token": xAccessToken
          }
        });

        firstWalletId = JSON.parse(allWallets.data.objectBytes)[0].walletId;
      } catch (error) {
        logger.error(error);
      }
    });

    it("Admin can mint tokens", async function() {
      let firstWallet = await axios({
        method: "get",
        url: `http://localhost:3000/wallets/${firstWalletId}`,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      let oldBalance = JSON.parse(firstWallet.data.objectBytes).Balance;
      let toMint = 999;
      await axios({
        method: "put",
        url: `http://localhost:3000/wallets/${firstWalletId}/mint`,
        data: { Amount: toMint },
        json: true,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      firstWallet = await axios({
        method: "get",
        url: `http://localhost:3000/wallets/${firstWalletId}`,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      let newBalance = JSON.parse(firstWallet.data.objectBytes).Balance;
      expect(newBalance - toMint).to.be.equal(oldBalance);
    });

    it("Admin can burn tokens", async function() {
      let firstWallet = await axios({
        method: "get",
        url: `http://localhost:3000/wallets/${firstWalletId}`,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      let oldBalance = JSON.parse(firstWallet.data.objectBytes).Balance;
      let toBurn = 999;
      await axios({
        method: "put",
        url: `http://localhost:3000/wallets/${firstWalletId}/burn`,
        data: { Amount: toBurn },
        json: true,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      firstWallet = await axios({
        method: "get",
        url: `http://localhost:3000/wallets/${firstWalletId}`,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      let newBalance = JSON.parse(firstWallet.data.objectBytes).Balance;

      expect(oldBalance - toBurn).to.be.equal(newBalance);
    });
  });

  describe("Non-admin actions", async function() {
    before(async () => {
      try {
        const email = `function${nanoid()}@fakeemail.com;`;
        await axios({
          method: "post",
          url: "http://localhost:3000/users/register",
          data: {
            email,
            password: "p@ssw0rd",
            org: "org1",
            role: "consumer"
          },
          headers: {
            "content-type": "application/json"
          }
        });
        let response = await axios({
          method: "post",
          url: "http://localhost:3000/users/login",
          data: {
            email,
            password: "p@ssw0rd",
            org: "org1"
          },
          headers: {
            "content-type": "application/json"
          }
        });
        xAccessToken = response.data.accessToken;
        allWallets = await axios({
          method: "get",
          url: "http://localhost:3000/wallets",
          headers: {
            "content-type": "application/json",
            "x-access-token": xAccessToken
          }
        });

        firstWalletId = JSON.parse(allWallets.data.objectBytes)[0].walletId;
      } catch (error) {
        logger.error(error);
      }
    });

    it("Non-admin cannot mint tokens", async function() {
      let firstWallet = await axios({
        method: "get",
        url: `http://localhost:3000/wallets/${firstWalletId}`,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      let oldBalance = JSON.parse(firstWallet.data.objectBytes).Balance;
      let toMint = 999;

      await expect(
        axios({
          method: "put",
          url: `http://localhost:3000/wallets/${firstWalletId}/mint`,
          data: { Amount: toMint },
          json: true,
          headers: {
            "content-type": "application/json",
            "x-access-token": xAccessToken
          }
        })
      ).to.be.rejectedWith(Error);
    });

    it("Non-admin cannot burn tokens", async function() {
      let firstWallet = await axios({
        method: "get",
        url: `http://localhost:3000/wallets/${firstWalletId}`,
        headers: {
          "content-type": "application/json",
          "x-access-token": xAccessToken
        }
      });

      let oldBalance = JSON.parse(firstWallet.data.objectBytes).Balance;
      let toBurn = 999;

      await expect(
        axios({
          method: "put",
          url: `http://localhost:3000/wallets/${firstWalletId}/burn`,
          data: { Amount: toBurn },
          json: true,
          headers: {
            "content-type": "application/json",
            "x-access-token": xAccessToken
          }
        })
      ).to.be.rejectedWith(Error);
    });
  });
});
